@@ -47,8 +47,12 @@
 #include <endian.h>
 #include <signal.h>
 
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_SYS_XATTR_H
+#include <sys/xattr.h>
+#define HAVE_XATTR 1
+#elif defined(HAVE_ATTR_XATTR_H)
 #include <attr/xattr.h>
+#define HAVE_XATTR 1
 #endif
 
 #if __WORDSIZE == 64
@@ -90,7 +94,7 @@ typedef struct _crypt_state {
 	size_t saltlen;
 	loff_t lastpos;
 	loff_t processed;
-#if 1 //def HAVE_ATTR_XATTR_H
+#if 1 //def HAVE_XATTR
 	char* salt_xattr_name;
 	char sxattr, sxfallback;
 	char* key_xattr_name;
@@ -103,6 +107,7 @@ typedef struct _crypt_state {
 	char outkeyiv;
 	char ctrbug198;
 	char opbkdf11;
+	char nosalthdr;
 } crypt_state;
 
 /* aes modules rely on avail of global crypto symbol to point to sec_fields ... */
@@ -113,22 +118,19 @@ const char *crypt_help = "The crypt plugin for dd_rescue de/encrypts data copied
 		" Parameters: [alg[o[rithm]]=]ALG:enc[rypt]:dec[rypt]:engine=STR:pad=STR\n"
 		"\t:keyhex=HEX:keyfd=[x]INT[@INT@INT]:keyfile=NAME[@INT@INT]:keygen:keysfile\n"
 		"\t:ivhex=HEX:ivfd=[x]INT[@INT@INT]:ivfile=NAME[@INT@INT]:ivgen:ivsfile\n"
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		"\t:keyxattr[=xattr_name]:kxfallback:ivxattr[=xattr_name]:ixfallback\n"
 #endif
 		"\t:pass=STR:passfd=[x]INT[@INT@INT]:passfile=NAME[@INT@INT]\n"
 		"\t:salt=STR:salthex=HEX:saltfd=[x]INT[@INT@INT]:saltfile=NAME[@INT@INT]\n"
 		"\t:saltlen=INT:saltgen:saltsfile"
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		":saltxattr[=xattr_name]:sxfallback"
 #endif
-		"\n\t:pbkdf2[=INT]:opbkdf[11]:debug:bench[mark]:skiphole:weakrnd:outkeyiv:ctrbug198\n"
+		"\n\t:pbkdf2[=INT]:opbkdf[11]:debug:bench[mark]:skiphole:weakrnd:outkeyiv\n"
+		"\t:nosalthdr:ctrbug198\n"
 		" Use algorithm=help to get a list of supported crypt algorithms\n";
 
-/* TODO: 
- * openssl compatibility (Salted__ <SALT> header)
- */
-
 int parse_hex(unsigned char*, const char*, uint maxlen);
 int parse_hex_u32(unsigned int*, const char*, uint maxlen);
 int read_fd(unsigned char*, const char*, uint maxlen, const char*);
@@ -158,7 +160,7 @@ int set_alg(crypt_state* state, const char* algnm)
 			FPLOG(FATAL, "Don't understand option (alg?) %s\n", algnm);
 		return -1;
 	}
-	if (!strcmp(algnm, "help")) {
+	if (!strcasecmp(algnm, "help")) {
 		FPLOG(INFO, "Crypto algorithms:", NULL);
 		ciph_desc_t *alg;
 		for (alg = state->engine; alg->name != NULL; ++alg)
@@ -175,7 +177,7 @@ int set_alg(crypt_state* state, const char* algnm)
 	return 0;
 }
 
-#define BLKSZ (state->alg? state->alg->blocksize: 16)
+#define BLKSZ (state->alg? state->alg->blocksize: 16UL)
 
 int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 {
@@ -192,10 +194,23 @@ int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 	state->pad = PAD_ALWAYS;
 	state->saltlen = -1;
 #if defined(CC_FLAGS_AES) && (defined(__x86_64__) || defined(__i386__))
-	if (have_aesni) { 
-		state->engine = AESNI_Methods;
+	if (have_aesni) {
+#ifndef NO_AVX2
+		if (have_avx2) {
+			state->engine = VAESNI_Methods;
+			if (opt->verbose)
+				FPLOG(INFO, "Default to use AVX2 AESNI crypto extensions\n");
+		} else {
+#else
+		do {
+#endif
+		state->engine = SAESNI_Methods;
 		if (opt->verbose)
-			FPLOG(INFO, "Default to use AESNI crypto extensions\n");
+			FPLOG(INFO, "Default to use Std AESNI crypto extensions\n");
+		}
+#ifdef NO_AVX2
+		while(0);
+#endif
 	} else {
 #elif defined(HAVE_AES_ARM64)
 	if (have_arm8crypto) {
@@ -218,28 +233,36 @@ int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			param = next;
 			continue;
 		}
-		if (!strcmp(param, "help")) {
+		if (!strcasecmp(param, "help")) {
 			FPLOG(INFO, "%s", crypt_help);
 			return -1;
-		} else if (!strcmp(param, "debug"))
+		} else if (!strcasecmp(param, "debug"))
 			state->debug = 1;
 		else if (!strcmp(param, "encrypt") || !strcmp(param, "enc"))
 			state->enc = 1;
 		else if (!strcmp(param, "decrypt") || !strcmp(param, "dec"))
 			state->enc = 0;
 		else if (!memcmp(param, "engine=", 7)) {
-			if (!strcmp(param+7, "aes_c"))
+			if (!strcasecmp(param+7, "aes_c"))
 				state->engine = AES_C_Methods;
 #ifdef HAVE_AES_ARM64
-			else if (!strcmp(param+7, "aesarm64"))
+			else if (!strcasecmp(param+7, "aesarm64"))
 				state->engine = AES_ARM8_Methods;
 #endif
 #ifdef HAVE_AESNI
-			else if (!strcmp(param+7, "aesni"))
-				state->engine = AESNI_Methods;
+#ifndef NO_AVX2
+			else if (!strcasecmp(param+7, "aesni") && have_avx2)
+				state->engine = VAESNI_Methods;
+			else if (!strcasecmp(param+7, "vaesni"))
+				state->engine = VAESNI_Methods;
+#endif
+			else if (!strcasecmp(param+7, "saesni"))
+				state->engine = SAESNI_Methods;
+			else if (!strcasecmp(param+7, "aesni"))
+				state->engine = SAESNI_Methods;
 #endif
 #ifdef HAVE_LIBCRYPTO
-			else if (!strcmp(param+7, "openssl"))
+			else if (!strcasecmp(param+7, "openssl"))
 				state->engine = AES_OSSL_Methods;
 #endif
 			else {
@@ -257,11 +280,11 @@ int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 		else if (!memcmp(param, "alg=", 4))
 			err += set_alg(state, param+4);
 		else if (!memcmp(param, "pad=", 4)) {
-			if (!strcmp(param+4, "zero"))
+			if (!strcasecmp(param+4, "zero"))
 				state->pad = PAD_ZERO;
-			else if (!strcmp(param+4, "always"))
+			else if (!strcasecmp(param+4, "always"))
 				state->pad = PAD_ALWAYS;
-			else if (!strcmp(param+4, "asneeded"))
+			else if (!strcasecmp(param+4, "asneeded"))
 				state->pad = PAD_ASNEEDED;
 			else {
 				FPLOG(FATAL, "Illegal padding %s: Specify zero/always/asneeded!\n",
@@ -346,7 +369,7 @@ int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			state->saltlen = ATOL(param+8);
 		else if (!strcmp(param, "saltgen"))
 			state->sgen = 1;
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		else if (!strcmp(param, "saltxattr"))
 			err += set_flag(&state->sxattr, "saltxattr");
 		else if (!memcmp(param, "saltxattr=", 10)) {
@@ -381,6 +404,8 @@ int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			state->opbkdf = 1; state->opbkdf11 = 1;
 		} else if (!strcmp(param, "skiphole"))
 			state->skiphole = 1;
+		else if (!strcmp(param, "nosalthdr"))
+			state->nosalthdr = 1;
 		else if (!strcmp(param, "weakrnd"))
 			state->weakrnd = 1;
 		else if (!strcmp(param, "outkeyiv"))
@@ -401,8 +426,10 @@ int crypt_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 	}
 	/* 1st: Set engine: Default: aesni/aes_c: Done */
 	/* 2nd: Set alg: Already done if set explicitly */
-	if (!err && !state->alg)
+	if (!err && !state->alg) {
 		state->alg = findalg(state->engine, "AES192-CTR", 0);
+		FPLOG(INFO, "Using default algorithm AES192-CTR\n", NULL);
+	}
 	if (!state->alg)
 		return -1;
 
@@ -440,6 +467,14 @@ int crypt_plug_release(void **stat)
 		secmem_release(state->sec);
 	else
 		return -2;
+#ifdef HAVE_XATTR
+	if (state->iv_xattr_name)
+		free(state->iv_xattr_name);
+	if (state->key_xattr_name)
+		free(state->key_xattr_name);
+	if (state->salt_xattr_name)
+		free(state->salt_xattr_name);
+#endif
 	free(*stat);
 	return 0;
 }
@@ -709,7 +744,7 @@ int write_keyfile(crypt_state *state, const char* base, const char* name, const
 	return err;
 }
 
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 int get_xattr(crypt_state *state, const char* atrnm,
 		unsigned char* data, int dlen,
 		char fb, char* fbf, char* flag)
@@ -883,7 +918,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 	sprintf(ivsnm, "IVS.%s", state->alg->name);
 	sprintf(keynm, "KEYS.%s", state->alg->name);
 	sprintf(saltnm, "SALT.%s", state->alg->name);
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 	if (state->sxattr && !state->salt_xattr_name) {
 		state->salt_xattr_name = malloc(32);
 		snprintf(state->salt_xattr_name, 32, "user.salt.%s", state->alg->name);
@@ -943,22 +978,24 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 		}
 
 		/* 5b: Decoding with openssl salt: Try reading */
-		if (state->opbkdf && !state->enc) {
+		if (state->opbkdf && !state->enc && !state->nosalthdr) {
 			char buf[16];
 			int fd = open(opt->iname, O_RDONLY);
-			if (fd > 0 && read(fd, buf, 16) == 16)
+			if (fd > 0 && read(fd, buf, 16) == 16) {
 				if (!memcmp(buf, "Salted__", 8)) {
 					memcpy(state->sec->salt, buf+8, 8);
 					state->sset = 1;
 					((opt_t*)opt)->init_ipos += 16; 
 					((fstate_t*)fst)->ipos += 16;
-				}
+				} else
+					state->nosalthdr = 1;
+			}
 			if (fd > 0)
 				close(fd);
 		}
 
 		/* 5c */
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		/* Try getting salt from xattr */
 		if (!state->sset && state->sxattr && !get_salt_xattr(state) && !state->enc)
 			state->sxattr = 0;
@@ -999,7 +1036,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 			if (write_file(state->sec->salt, state->sfnm, 8, 0640))
 				return -1;
 		}
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		/* Write salt to xattr */
 		if (state->sxattr && state->enc && set_salt_xattr(state)) {
 			if (!state->sxfallback)
@@ -1028,7 +1065,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 			if (!state->keyf && !state->kxattr)
 				FPLOG(WARN, "Generated key not written anywhere?\n", NULL);
 			else {
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 				/* Write key to xattr, failure is fatal */
 				if (state->kxattr && state->enc && set_key_xattr(state) && !state->kxfallback)
 					return -1;
@@ -1066,7 +1103,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 				FPLOG(FATAL, "Key generation with pass+salt failed!\n", NULL);
 				return -1;
 			}
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 			/* Write key to xattr, failure is fatal */
 			if (state->kxattr && state->enc && set_key_xattr(state) && !state->kxfallback)
 				return -1;
@@ -1077,7 +1114,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 					return -1;
 
 		} else {
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 			if (state->kxattr)
 				get_key_xattr(state);
 #endif
@@ -1100,7 +1137,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 			}
 		}
 	} else {
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		if (state->kxattr && set_key_xattr(state) && !state->kxfallback) {
 			FPLOG(FATAL, "Can't save key in xattr");
 			return -1;
@@ -1123,7 +1160,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 			if (!state->ivf && !state->ixattr)
 				FPLOG(WARN, "Generated IV not saved?\n", NULL);
 			else {
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 				/* Write IV to xattr, failure w/o fb is fatal */
 				if (state->ixattr && state->enc && set_iv_xattr(state) && !state->ixfallback)
 					return -1;
@@ -1149,7 +1186,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 					return -1;
 				}
 			}
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 			/* Write IV to xattr, failure w/o fb is fatal */
 			if (state->ixattr && state->enc && set_iv_xattr(state) && !state->ixfallback)
 				return -1;
@@ -1159,7 +1196,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 				if (write_keyfile(state, ivsnm, encnm, state->sec->nonce1, BLKSZ, 0640, 1, 0))
 					return -1;
 		} else { 
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 			if (state->ixattr)
 				get_iv_xattr(state);
 #endif
@@ -1180,7 +1217,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 			}
 		 }
 	} else if (state->alg->stream->needs_iv) {
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		if (state->ixattr && set_iv_xattr(state) && !state->ixfallback) {
 			FPLOG(FATAL, "Can't save IV in xattr");
 			return -1;
@@ -1198,7 +1235,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 		FPLOG(DEBUG, " IV: %s\n", state->sec->charbuf1);
 	}
 	/* Write Salted__ header */
-	if (state->opbkdf && state->enc) {
+	if (state->opbkdf && state->enc && !state->nosalthdr) {
 		char buf[16];
 		strcpy(buf, "Salted__");
 		memcpy(buf+8, state->sec->salt, 8);
@@ -1223,7 +1260,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 	/* IV */
 	if (state->alg->stream->iv_prep)
 		state->alg->stream->iv_prep(state->sec->nonce1, state->sec->iv1.data,
-					    state->lastpos/BLKSZ-state->opbkdf);
+					    state->lastpos/BLKSZ-(state->opbkdf & !state->nosalthdr));
 	else
 		memcpy(state->sec->iv1.data, state->sec->nonce1, BLKSZ);
 
@@ -1234,7 +1271,7 @@ int crypt_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 	/* No need to keep key/passphr in memory */
 	memset(state->sec->userkey1, 0, state->alg->keylen/8);
 	memset(state->sec->passphr, 0, 128);
-	asm("":::"memory");
+	LFENCE;
 	if (state->bench)
 		state->cpu += clock() - t1;
 	return err;
@@ -1260,7 +1297,7 @@ char memcpy_testzero(void* dst, const void* src, size_t ln)
 			return 0;
 		}
 	}
-	asm volatile ("":::"memory");
+	LFENCE;
 	return 1;
 }
 
@@ -1309,25 +1346,27 @@ unsigned char* crypt_blk_cb(fstate_t *fst, unsigned char* bf,
 		}
 	}
 	if (0 && state->debug)
-		FPLOG(DEBUG, "%zi: %02x %02x %02x %02x ... -> ",
+		FPLOG(DEBUG, "%Li: %02x %02x %02x %02x ... -> ",
 			currpos, bf[0], bf[1], bf[2], bf[3]);
 	if (((currpos) % BLKSZ) && state->enc) {
-		FPLOG(WARN, "Enc alignment error! (%zi-%i)=%zi %i/%i\n", currpos, state->inbuf,
-			currpos - state->inbuf,
-			(currpos-state->inbuf)%BLKSZ, (currpos-state->inbuf)&0x0f);
+		const unsigned int curroff = (unsigned int)currpos % BLKSZ;
+		const unsigned int currmininoff = (unsigned int)(currpos-state->inbuf)%BLKSZ;
+		FPLOG(WARN, "Enc alignment error! (%Li-%i)=%Li %i/%i\n", currpos, state->inbuf,
+			currpos - state->inbuf, currmininoff,
+			(unsigned int)((currpos-state->inbuf)&0x0fUL));
 		/* Can only handle in CTR mode and without buffered bytes. */
 		assert(state->alg->stream->granul == 1);
 		assert(state->inbuf == 0);
-		memcpy(state->sec->databuf1+(currpos%BLKSZ), bf, BLKSZ-currpos%BLKSZ);
+		memcpy(state->sec->databuf1+curroff, bf, BLKSZ-curroff);
 		err = crypt(keys, state->alg->rounds, state->sec->iv1.data,
-			    PAD_ZERO, state->sec->databuf1, bf-(currpos%BLKSZ), BLKSZ, &olen);
+			    PAD_ZERO, state->sec->databuf1, bf-curroff, BLKSZ, &olen);
 		assert(!err);
 		assert(olen == BLKSZ);
-		i = BLKSZ-(currpos%BLKSZ);
+		i = BLKSZ-curroff;
 	} else if ((currpos-state->inbuf)%BLKSZ && !state->enc) {
-		FPLOG(WARN, "Dec alignment error! (%zi-%i)=%zi %i/%i\n", currpos, state->inbuf,
-			currpos - state->inbuf,
-			(currpos-state->inbuf)%BLKSZ, (currpos-state->inbuf)&0x0f);
+		FPLOG(WARN, "Dec alignment error! (%Li-%i)=%Li %i/%i\n", currpos, state->inbuf,
+			currpos - state->inbuf, (unsigned int)((currpos-state->inbuf)%BLKSZ),
+			(unsigned int)((currpos-state->inbuf)&0x0fUL));
 		//raise(SIGQUIT);
 	}
 	if (!state->enc && !state->rev)
@@ -1449,12 +1488,13 @@ int crypt_close(loff_t ooff, void **stat)
 	crypt_state *state = (crypt_state*)*stat;
 	assert(state->inbuf == 0);
 	state->alg->release(state->enc? state->sec->ekeys->data: state->sec->dkeys->data, state->alg->rounds);
-	/* secmem_release(state->sec) is calles in crypt_plug_release */
+	/* secmem_release(state->sec) is called in crypt_plug_release */
 	if (state->bench && state->cpu/(CLOCKS_PER_SEC/20) > 0)
 		FPLOG(INFO, "%.2fs CPU time, %.1fMiB/s\n",
 			(double)state->cpu/CLOCKS_PER_SEC, 
 			state->processed/1024 / (state->cpu/(CLOCKS_PER_SEC/1024.0)));
-#ifdef HAVE_ATTR_XATTR_H
+	/* xattr mem is released in release */
+#if 0 //def HAVE_XATTR
 	if (state->salt_xattr_name)
 		free(state->salt_xattr_name);
 	if (state->key_xattr_name)
