@@ -23,6 +23,7 @@
 #include "sha1.h"
 #include "pbkdf2.h"
 #include "checksum_file.h"
+#include "secmem.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -36,8 +37,12 @@
 #include <netinet/in.h>	/* For ntohl/htonl */
 #include <endian.h>
 
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_SYS_XATTR_H
+#include <sys/xattr.h>
+#define HAVE_XATTR 1
+#elif defined(HAVE_ATTR_XATTR_H)
 #include <attr/xattr.h>
+#define HAVE_XATTR 1
 #endif
 // TODO: pass at runtime rather than compile time
 #define HASH_DEBUG(x) if (state->debug) x
@@ -62,7 +67,7 @@ typedef struct _hash_state {
 	int seq;
 	int outfd;
 	unsigned char buflen;
-	unsigned char ilnchg, olnchg, ichg, ochg, debug, outf, chkf, chkfalloc;
+	unsigned char ilnchg, olnchg, ichg, ochg, debug, outf, chkf, chkfalloc, chkupd;
 	const char* chkfnm;
 	const opt_t *opts;
 	unsigned char* hmacpwd;
@@ -75,7 +80,7 @@ typedef struct _hash_state {
 #endif
 	int hmacpln;
 	char xfallback;
-#if 1 //def HAVE_ATTR_XATTR_H
+#if 1 //def HAVE_XATTR
 	char chk_xattr, set_xattr, xnmalloc;
 	char* xattr_name;
 #endif
@@ -83,14 +88,14 @@ typedef struct _hash_state {
 
 const char *hash_help = "The HASH plugin for dd_rescue calculates a cryptographic checksum on the fly.\n"
 		" It supports unaligned blocks (arbitrary offsets) and holes(sparse writing).\n"
-		" Parameters: output:outfd=FNO:outnm=FILE:check:chknm=FILE:debug:[alg[o[rithm]=]ALG\n"
-		"\t:append=STR:prepend=STR:hmacpwd=STR:hmacpwdfd=FNO:hmacpwdnm=FILE\n"
+		" Parameters: output:outfd=FNO:outnm=FILE:check:chknm=FILE:chkadd:[alg[o[rithm]=]ALG\n"
+		"\t:append=STR:prepend=STR:hmacpwd=STR:hmacpwdfd=FNO:hmacpwdnm=FILE:debug\n"
 		"\t:pbkdf2=ALG/PWD/SALT/ITER/LEN"
 #ifndef NO_S3_MP
 		":multipart=size"
 #endif
 		"\n"
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		"\t:chk_xattr[=xattr_name]:set_xattr[=xattr_name]:fallb[ack][=FILE]\n"
 #endif
 		" Use algorithm=help to get a list of supported hash algorithms\n";
@@ -120,7 +125,7 @@ static loff_t readint(const char* const ptr)
 hashalg_t *get_hashalg(hash_state *state, const char* nm)
 {
 	unsigned int i;
-	const char help = !strcmp(nm, "help");
+	const char help = !strcasecmp(nm, "help");
 	if (help)
 		FPLOG(INFO, "Supported algorithms:");
 	for (i = 0; i < sizeof(hashes)/sizeof(hashalg_t); ++i) {
@@ -142,7 +147,7 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 	int err = 0;
 	hash_state *state; /* = (hash_state*)malloc(sizeof(hash_state));*/
 	if (posix_memalign(stat, 64, sizeof(hash_state))) {
-		FPLOG(FATAL, "No enough memory for hash state!\n");
+		FPLOG(FATAL, "Not enough memory for hash state!\n");
 		return -1;
 	}
 	state = (hash_state*)*stat;
@@ -154,10 +159,10 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 		char* next = strchr(param, ':');
 		if (next)
 			*next++ = 0;
-		if (!strcmp(param, "help"))
+		if (!strcasecmp(param, "help"))
 			FPLOG(INFO, "%s", hash_help);
 		
-		else if (!strcmp(param, "debug"))
+		else if (!strcasecmp(param, "debug"))
 			state->debug = 1;
 		else if (!strcmp(param, "output"))
 			state->outfd = 1;
@@ -167,7 +172,7 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			state->append = param+7;
 		else if (!memcmp(param, "prepend=", 8))
 			state->prepend = param+8;
-#if 1 //def HAVE_ATTR_XATTR_H
+#if 1 //def HAVE_XATTR
 		else if (!memcmp(param, "chk_xattr=", 10)) {
 			state->chk_xattr = 1; state->xattr_name = param+10; }
 		else if (!strcmp(param, "chk_xattr"))
@@ -193,6 +198,8 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			state->chkf = 1;
 		else if (!memcmp(param, "chknm=", 6)) {
 			state->chkf = 1; state->chkfnm=param+6; }
+		else if (!memcmp(param, "chkadd", 6))
+			state->chkupd = 1;
 		else if (!strcmp(param, "check")) {
 			state->chkf = 1; state->chkfnm="-"; }
 #ifndef NO_S3_MP
@@ -207,17 +214,23 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			state->alg = get_hashalg(state, param+10);
 		else if (!memcmp(param, "hmacpwd=", 8)) {
 			state->hmacpwd = (unsigned char*)malloc(MAX_HMACPWDLN);
-			strncpy((char*)state->hmacpwd, param+8, MAX_HMACPWDLN);
+			assert(state->hmacpwd);
 			state->hmacpln = strlen(param+8);
+			if (state->hmacpln >= MAX_HMACPWDLN)
+				state->hmacpln = MAX_HMACPWDLN-1;
+			memcpy(state->hmacpwd, param+8, state->hmacpln);
+			state->hmacpwd[state->hmacpln] = 0;
 		}
 		else if (!memcmp(param, "hmacpwdfd=", 10)) {
 			int hfd = atol(param+10);
 			state->hmacpwd = (unsigned char*)malloc(MAX_HMACPWDLN);
+			assert(state->hmacpwd);
+			state->hmacpwd[MAX_HMACPWDLN-1] = 0;
 			if (hfd == 0 && isatty(hfd)) {
 				FPLOG(INPUT, "%s", "Enter HMAC password: ");
-				state->hmacpln = hidden_input(hfd, (char*)state->hmacpwd, MAX_HMACPWDLN, 1);
+				state->hmacpln = hidden_input(hfd, (char*)state->hmacpwd, MAX_HMACPWDLN-1, 1);
 			} else
-				state->hmacpln = read(hfd, state->hmacpwd, MAX_HMACPWDLN);
+				state->hmacpln = read(hfd, state->hmacpwd, MAX_HMACPWDLN-1);
 			if (state->hmacpln <= 0) {
 				FPLOG(FATAL, "No HMAC password entered!\n");
 				--err;
@@ -232,7 +245,9 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 				continue;
 			}
 			state->hmacpwd = (unsigned char*)malloc(MAX_HMACPWDLN);
-			state->hmacpln = fread(state->hmacpwd, 1, MAX_HMACPWDLN, f);
+			assert(state->hmacpwd);
+			state->hmacpwd[MAX_HMACPWDLN-1] = 0;
+			state->hmacpln = fread(state->hmacpwd, 1, MAX_HMACPWDLN-1, f);
 			if (state->hmacpln <= 0) {
 				FPLOG(FATAL, "No HMAC pwd contents found in %s!\n", param+10);
 				--err;
@@ -258,9 +273,10 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 		FPLOG(FATAL, "No hash algorithm specified\n");
 		return --err;
 	}
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 	if ((state->chk_xattr || state->set_xattr) && !state->xattr_name) {
 		state->xattr_name = (char*)malloc(32);
+		assert(state->xattr_name);
 		state->xnmalloc = 1;
 		if (state->hmacpwd)
 			snprintf(state->xattr_name, 32, "user.hmac.%s", state->alg->name);
@@ -269,7 +285,7 @@ int hash_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 	}
 #endif
 	if ((!state->chkfnm || !*state->chkfnm) && (state->chkf || state->outf
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 				|| state->xfallback
 #endif
 			     				)) {
@@ -301,7 +317,7 @@ int hash_plug_release(void **stat)
 	if (!stat || !(*stat))
 		return -1;
 	hash_state *state = (hash_state*)*stat;
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 	if (state->xnmalloc)
 		free((void*)state->xattr_name);
 #endif
@@ -311,7 +327,7 @@ int hash_plug_release(void **stat)
 		free((void*)state->fname);
 	if (state->hmacpwd) {
 		memset(state->hmacpwd, 0, MAX_HMACPWDLN);
-		asm("":::"memory");
+		LFENCE;
 		free(state->hmacpwd);
 	}
 	free(*stat);
@@ -351,7 +367,7 @@ int hash_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 		strcat(nnm, "->");
 		strcat(nnm, opt->oname);
 		state->fname = nnm;
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 		if (state->chk_xattr || state->set_xattr) {
 			--err;
 			FPLOG(WARN, "Can't access xattr in the middle of a plugin chain!");
@@ -564,31 +580,6 @@ unsigned char* hash_blk_cb(fstate_t *fst, unsigned char* bf,
 }
 
 
-int check_chkf(hash_state *state, const char* res)
-{
-	const char* name = state->opts->iname;
-	char cks[144];
-	if (state->ichg && !state->ochg) {
-		name = state->opts->oname;
-		if (!state->opts->quiet)
-			FPLOG(INFO, "Read checksum from %s for output file %s\n", state->chkfnm, name);
-	} else if (state->ichg) {
-		FPLOG(WARN, "Can't read checksum in the middle of plugin chain (%s)\n", state->fname);
-		return -ENOENT;
-	}
-	int err = get_chks(state->chkfnm, name, cks, strlen(res));
-	if (err < 0) {
-		FPLOG(WARN, "Can't find checksum in %s for %s\n", state->chkfnm, name);
-		return -ENOENT;
-	}
-	if (strcmp(cks, res)) {
-		FPLOG(WARN, "Hash from chksum file %s for %s does not match\n", state->chkfnm, name);
-		FPLOG(WARN, "comp %s, read %s\n", res, cks);
-		return -EBADF;
-	}
-	return 0;
-}
-
 int write_chkf(hash_state *state, const char *res)
 {
 	const char* name = state->opts->oname;
@@ -607,48 +598,36 @@ int write_chkf(hash_state *state, const char *res)
 	return err;
 }
 
-#ifdef HAVE_ATTR_XATTR_H
-int check_xattr(hash_state* state, const char* res)
+int check_chkf(hash_state *state, const char* res)
 {
-	char xatstr[144];
-	strcpy(xatstr, "xattr");
 	const char* name = state->opts->iname;
+	char cks[144];
 	if (state->ichg && !state->ochg) {
 		name = state->opts->oname;
 		if (!state->opts->quiet)
-			FPLOG(INFO, "Read xattr from output file %s\n", name);
+			FPLOG(INFO, "Read checksum from %s for output file %s\n", state->chkfnm, name);
 	} else if (state->ichg) {
-		FPLOG(WARN, "Can't read xattrs in the middle of plugin chain (%s)\n", state->fname);
+		FPLOG(WARN, "Can't read checksum in the middle of plugin chain (%s)\n", state->fname);
 		return -ENOENT;
 	}
-	/* Longest is 128byte hex for SHA512 (8x64byte numbers -> 8x16 digits) */
-	char chksum[144];
-	ssize_t itln = getxattr(name, state->xattr_name, chksum, 144);
-	const int rln = strlen(res);
-	if (itln <= 0) {
-		if (state->xfallback) {
-			int err = get_chks(state->chkfnm, name, chksum, rln);
-			snprintf(xatstr, 143, "chksum file %s", state->chkfnm);
-			if (err < 0) {
-				FPLOG(WARN, "no hash found in xattr nor %s for %s\n", xatstr, name);
-				return -ENOENT;
-			} else if (strcmp(chksum, res)) {
-				FPLOG(WARN, "Hash from %s for %s does not match\n", xatstr, name);
-				return -EBADF;
-			}
-		} else {
-			FPLOG(WARN, "Hash could not be read from xattr of %s\n", name);
+	int err = get_chks(state->chkfnm, name, cks, strlen(res));
+	if (err < 0) {
+		if (state->chkupd)
+			return write_chkf(state, res);
+		else {
+			FPLOG(WARN, "Can't find checksum in %s for %s\n", state->chkfnm, name);
 			return -ENOENT;
 		}
-	} else if (itln < rln || memcmp(res, chksum, rln)) {
-		FPLOG(WARN, "Hash from xattr of %s does not match\n", name);
+	}
+	if (strcmp(cks, res)) {
+		FPLOG(WARN, "Hash from chksum file %s for %s does not match\n", state->chkfnm, name);
+		FPLOG(WARN, "comp %s, read %s\n", res, cks);
 		return -EBADF;
 	}
-	if (!state->opts->quiet || state->debug)
-		FPLOG(INFO, "Successfully validated hash from %s for %s\n", xatstr, name);
 	return 0;
 }
 
+#ifdef HAVE_XATTR
 int write_xattr(hash_state* state, const char* res)
 {
 	const char* name = state->opts->oname;
@@ -682,6 +661,51 @@ int write_xattr(hash_state* state, const char* res)
 				xatstr, name, res);
 	return 0;
 }
+
+int check_xattr(hash_state* state, const char* res)
+{
+	char xatstr[144];
+	strcpy(xatstr, "xattr");
+	const char* name = state->opts->iname;
+	if (state->ichg && !state->ochg) {
+		name = state->opts->oname;
+		if (!state->opts->quiet)
+			FPLOG(INFO, "Read xattr from output file %s\n", name);
+	} else if (state->ichg) {
+		FPLOG(WARN, "Can't read xattrs in the middle of plugin chain (%s)\n", state->fname);
+		return -ENOENT;
+	}
+	/* Longest is 128byte hex for SHA512 (8x64byte numbers -> 8x16 digits) */
+	char chksum[144];
+	ssize_t itln = getxattr(name, state->xattr_name, chksum, 144);
+	const int rln = strlen(res);
+	if (itln <= 0) {
+		if (state->xfallback) {
+			int err = get_chks(state->chkfnm, name, chksum, rln);
+			snprintf(xatstr, 143, "chksum file %s", state->chkfnm);
+			if (err < 0) {
+				if (state->chkupd)
+					return write_xattr(state, res);
+				else {
+					FPLOG(WARN, "no hash found in xattr nor %s for %s\n", xatstr, name);
+					return -ENOENT;
+				}
+			} else if (strcmp(chksum, res)) {
+				FPLOG(WARN, "Hash from %s for %s does not match\n", xatstr, name);
+				return -EBADF;
+			}
+		} else {
+			FPLOG(WARN, "Hash could not be read from xattr of %s\n", name);
+			return -ENOENT;
+		}
+	} else if (itln < rln || memcmp(res, chksum, rln)) {
+		FPLOG(WARN, "Hash from xattr of %s does not match\n", name);
+		return -EBADF;
+	}
+	if (!state->opts->quiet || state->debug)
+		FPLOG(INFO, "Successfully validated hash from %s for %s\n", xatstr, name);
+	return 0;
+}
 #endif
 
 int hash_close(loff_t ooff, void **stat)
@@ -714,7 +738,7 @@ int hash_close(loff_t ooff, void **stat)
 		state->alg->hash_beout(obuf+blen, &state->hmach);
 		state->alg->hash_init(&state->hmach);
 		state->alg->hash_calc(obuf, blen+hlen, blen+hlen, &state->hmach);
-		memset(obuf, 0, blen); asm("":::"memory");
+		memset(obuf, 0, blen); LFENCE;
 		state->alg->hash_hexout(res, &state->hmach);
 		if (!state->opts->quiet) 
 			FPLOG(INFO, "HMAC %s %s (%" LL "i-%" LL "i): %s\n",
@@ -734,7 +758,7 @@ int hash_close(loff_t ooff, void **stat)
 		err += check_chkf(state, res);
 	if (state->outf)
 		err += write_chkf(state, res);
-#ifdef HAVE_ATTR_XATTR_H
+#ifdef HAVE_XATTR
 	if (state->chk_xattr)
 		err += check_xattr(state, res);
 	if (state->set_xattr)
