@@ -10,8 +10,9 @@
 # include "config.h"
 #endif
 
-typedef struct _roundkey {
+typedef union _roundkey {
 	unsigned char data[16];
+	unsigned int data32[4];
 } roundkey;
 
 typedef struct _ciphblk {
@@ -40,15 +41,21 @@ typedef struct _sec_fields {
 	unsigned char salt[64];
 	unsigned char passphr[128];
 	/* @2816: char buffer, enough for 512bit hash/key in hex */
-	char charbuf1[144];
-	/* @2960: one block */
-	unsigned char blkbuf1[16];
-	/* @2976: data buffer for incomplete blocks */
-	unsigned char databuf1[32];
-	/* @3008: four blocks (64B) */
-	unsigned char blkbuf2[64];
+	char charbuf1[160];
+	/* @2976: two blocks - intermediate result for x2 cypto */
+	unsigned char blkbuf1[32];
+	/* @3008: data buffer for incomplete blocks (libcrypt) */
+	unsigned char databuf1[64];
 	/* @3072: buffer up to 512 bytes */
 	unsigned char databuf2[512];
+	/* @3584: four blocks (64B) - incomplete input block */
+	unsigned char blkbuf2[64];
+	/* @3648: four blocks (64B) - output buffer */
+	unsigned char blkbuf3[64];
+	/* @3712: four blocks (64B) - x2 long buffer */
+	unsigned char blkbuf4[64];
+	/* @3776 canary */
+	unsigned long long canary;
 	
 } sec_fields;
 
@@ -57,4 +64,12 @@ void secmem_release(sec_fields*);
 
 extern sec_fields *crypto;
 
+//#if defined(__i386__) || defined(__x86_64__)
+#ifdef HAVE_LFENCE
+#define LFENCE asm("lfence":::"memory")
+#else
+#define LFENCE asm("":::"memory")
+#endif
+#define mem_clobber asm("":::"memory")
+
 #endif
